// -*- mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:t -*-
// vim: ts=8 sw=8 smarttab
#include "graid.h"
#include <linux/version.h>

#if defined(RHEL_DRM_VERSION_CODE)
#if RHEL_DRM_VERSION_CODE < KERNEL_VERSION(5, 12, 0)
#define USE_OLD_VM_MAP_RAM
#endif
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 8, 0)
#define USE_OLD_VM_MAP_RAM
#endif
#endif

struct ppptr {
	union {
		struct {
			atomic64_t kptr;
			atomic64_t pptr;
		};
		u8 pad[64];
	};
};
struct GRAID_Page_Pool {
	struct ppptr *ptrs;
	u8 pad1[56];
	atomic_t head;
	u8 pad2[60];
	atomic_t tail;
	u8 pad3[60];
	atomic_t free;
	u8 pad4[60];
};
struct PageResources {
	struct GRAID_DMA_Pool dma_pool;
	uint32_t page_pool_size;
	uint32_t page_pool_mask;
	struct page **kern_page_lists[GRAID_MAX_NUMA];
	struct page **gpu_page_lists[GRAID_MAX_NUMA];
	struct GRAID_Page_Pool page_pools[GRAID_MAX_NUMA];

	uint32_t gpu_page_ppn;
	uint32_t gpu_page_pool_size;
	uint32_t gpu_page_pool_mask;
	struct GRAID_Page_Pool gpu_page_pools[GRAID_MAX_NUMA];
};
static struct PageResources *res;

static struct graid_mem_addrs *graid_mem_kaddrs = NULL;
static struct graid_mem_addrs *graid_mem_paddrs = NULL;
static struct graid_mem_addrs *graid_bar1_kaddrs = NULL;
static struct graid_mem_addrs *graid_bar1_paddrs = NULL;
static int bar1_mapped = 0;

int graid_is_dma_pool(uint64_t addr, uint64_t len)
{
	int i;

	if (!res)
		return 0;

	/* DMA Page Pool Passes Kernel Address */
	for (i = 0; i < res->dma_pool.numa_nr; ++i) {
		int node = res->dma_pool.numa_nodes[i];
		if (addr == (u64)res->dma_pool.kern_page_base[node] &&
		    len  == ((u64)res->dma_pool.kern_alloc_pgnr) * PAGE_SIZE)
			return GRAID_DMA_NODE_OFFSET + node;
		if (addr == (u64)res->dma_pool.gpu_page_base[node] &&
		    len  == ((u64)res->dma_pool.gpu_alloc_pgnr) * PAGE_SIZE)
			return GRAID_DMA_NODE_OFFSET + node;
	}

	if (len == PAGE_SIZE) {
		if (graid_mem_paddrs->ipc == addr)
			return 1;
		for (i = 0; i < DB_COUNT; ++i)
			if (graid_mem_paddrs->dbs[i] == addr)
				return 1;
	}
	if (len == PDSQBytes) {
		for (i = 0; i < PDQ_COUNT; ++i)
			if (graid_mem_paddrs->pd_sqs[i] == addr)
				return 1;
	}
	if (len == PDCQBytes) {
		for (i = 0; i < PDQ_COUNT; ++i)
			if (graid_mem_paddrs->pd_cqs[i] == addr)
				return 1;
	}
	if (len == DGSQBytes) {
		for (i = 0; i < DGQ_COUNT; ++i)
			if (graid_mem_paddrs->dg_sqs[i] == addr)
				return 1;
	}
	if (len == DGCQBytes) {
		for (i = 0; i < DGQ_COUNT; ++i)
			if (graid_mem_paddrs->dg_cqs[i] == addr)
				return 1;
	}

	return 0;
}

dma_addr_t graid_pci_get_cq_paddr(int instance, int qid)
{
	int idx = instance * NVMeMaxPDQNR + qid - 1;
	BUG_ON(!qid);
	return (dma_addr_t) (graid_bar1_paddrs->pd_cqs[idx] ?:
			     graid_mem_paddrs->pd_cqs[idx]);
}

dma_addr_t graid_pci_get_sq_paddr(int instance, int qid)
{
	int idx = instance * NVMeMaxPDQNR + qid - 1;
	BUG_ON(!qid);
	return (dma_addr_t) (graid_bar1_paddrs->pd_sqs[idx] ?:
			     graid_mem_paddrs->pd_sqs[idx]);
}

dma_addr_t graid_vir_get_cq_paddr(int dg_id, int qid)
{
	int idx = (dg_id * NVMeMaxDGQNR) + qid - 1;
	BUG_ON(!qid);
	return (dma_addr_t) (graid_bar1_paddrs->dg_cqs[idx] ?:
			     graid_mem_paddrs->dg_cqs[idx]);
}

dma_addr_t graid_vir_get_sq_paddr(int dg_id, int qid)
{
	int idx = (dg_id * NVMeMaxDGQNR) + qid - 1;
	BUG_ON(!qid);
	return (dma_addr_t) (graid_bar1_paddrs->dg_sqs[idx] ?:
			     graid_mem_paddrs->dg_sqs[idx]);
}

void *graid_pci_get_cq_vaddr(int instance, int qid)
{
	int idx = instance * NVMeMaxPDQNR + qid - 1;
	BUG_ON(!qid);
	return (void*) (graid_bar1_kaddrs->pd_cqs[idx] ?:
			graid_mem_kaddrs->pd_cqs[idx]);
}

void *graid_pci_get_sq_vaddr(int instance, int qid)
{
	int idx = instance * NVMeMaxPDQNR + qid - 1;
	BUG_ON(!qid);
	return (void*) (graid_bar1_kaddrs->pd_sqs[idx] ?:
			graid_mem_kaddrs->pd_sqs[idx]);
}

void *graid_vir_get_cq_vaddr(int dg_id, int qid)
{
	int idx = (dg_id * NVMeMaxDGQNR) + qid - 1;
	BUG_ON(!qid);
	return (void*) (graid_bar1_kaddrs->dg_cqs[idx] ?:
			graid_mem_kaddrs->dg_cqs[idx]);
}

void *graid_vir_get_sq_vaddr(int dg_id, int qid)
{
	int idx = (dg_id * NVMeMaxDGQNR) + qid - 1;
	BUG_ON(!qid);
	return (void*) (graid_bar1_kaddrs->dg_sqs[idx] ?:
			graid_mem_kaddrs->dg_sqs[idx]);
}

void *graid_blk_get_dbs_vaddr(int pd_id)
{
	return (void*) (graid_bar1_kaddrs->dbs[pd_id] ?:
			graid_mem_kaddrs->dbs[pd_id]);
}

void *graid_vir_get_dbs_vaddr(int dg_id)
{
	return (void*) (graid_bar1_kaddrs->dbs[NVMeMaxPDNR + dg_id] ?:
			graid_mem_kaddrs->dbs[NVMeMaxPDNR + dg_id]);
}

static int alloc_dma_memory(struct device *dev, uint64_t *kaddrs,
			    uint64_t *daddrs, int size, int count,
			    int (*idx_to_node)(int))
{
	int i;

	for (i = 0; i < count; ++i) {
		void *kaddr;
		if (!idx_to_node) {
			kaddr = kmalloc(size, GFP_KERNEL);
		} else {
			int node = idx_to_node(i);
			if (node < 0)
				kaddr = kmalloc(size, GFP_KERNEL);
			else
				kaddr = kmalloc_node(size, GFP_KERNEL, node);
		}
		if (!kaddr)
			goto err_out;
		kaddrs[i] = (uint64_t)kaddr;
		daddrs[i] = virt_to_phys(kaddr);
	}
	return 0;

err_out:
	for (; i >=0; --i) {
		kfree((void*)kaddrs[i]);
		kaddrs[i] = 0;
		daddrs[i] = 0;
	}
	return -ENOMEM;
}

static void free_dma_memory(struct device *dev, uint64_t *kaddrs,
			    uint64_t *daddrs, int size, int count)
{
	int i;

	for (i = 0; i < count; ++i) {
		if (kaddrs[i]) {
			kfree((void*)kaddrs[i]);
			kaddrs[i] = 0;
			daddrs[i] = 0;
		}
	}
}

static int graid_alloc_queue_buffers(struct device *dev)
{
	int rc;

	graid_mem_kaddrs = vzalloc(sizeof(struct graid_mem_addrs));
	if (!graid_mem_kaddrs)
		goto out_free;

	graid_mem_paddrs = vzalloc(sizeof(struct graid_mem_addrs));
	if (!graid_mem_paddrs)
		goto out_free;

	graid_bar1_kaddrs = vzalloc(sizeof(struct graid_mem_addrs));
	if (!graid_bar1_kaddrs)
		goto out_free;

	graid_bar1_paddrs = vzalloc(sizeof(struct graid_mem_addrs));
	if (!graid_bar1_paddrs)
		goto out_free;

	graid_mem_kaddrs->ipc = (uint64_t)kmalloc(PAGE_SIZE, GFP_KERNEL);
	if (!graid_mem_kaddrs->ipc)
		goto out_free;
	graid_mem_paddrs->ipc = virt_to_phys((void*)graid_mem_kaddrs->ipc);

	rc = alloc_dma_memory(dev, graid_mem_kaddrs->pd_sqs, graid_mem_paddrs->pd_sqs,
			      PDSQBytes, PDQ_COUNT, NULL);
	if (rc < 0)
		goto out_free_ipc;

	rc = alloc_dma_memory(dev, graid_mem_kaddrs->pd_cqs, graid_mem_paddrs->pd_cqs,
			      PDCQBytes, PDQ_COUNT, NULL);
	if (rc < 0)
		goto out_free_pd_sq;

	rc = alloc_dma_memory(dev, graid_mem_kaddrs->dg_sqs, graid_mem_paddrs->dg_sqs,
			      DGSQBytes, DGQ_COUNT, NULL);
	if (rc < 0)
		goto out_free_pd_cq;

	rc = alloc_dma_memory(dev, graid_mem_kaddrs->dg_cqs, graid_mem_paddrs->dg_cqs,
			      DGCQBytes, DGQ_COUNT, graid_dgq_node);
	if (rc < 0)
		goto out_free_dg_sq;

	rc = alloc_dma_memory(dev, graid_mem_kaddrs->dbs, graid_mem_paddrs->dbs,
			      PAGE_SIZE, DB_COUNT, NULL);
	if (rc < 0)
		goto out_free_dg_cq;

	return 0;

out_free_dg_cq:
	free_dma_memory(dev, graid_mem_kaddrs->dg_cqs, graid_mem_paddrs->dg_cqs,
			DGCQBytes, DGQ_COUNT);
out_free_dg_sq:
	free_dma_memory(dev, graid_mem_kaddrs->dg_sqs, graid_mem_paddrs->dg_sqs,
			DGSQBytes, DGQ_COUNT);
out_free_pd_cq:
	free_dma_memory(dev, graid_mem_kaddrs->pd_cqs, graid_mem_paddrs->pd_cqs,
			PDCQBytes, PDQ_COUNT);
out_free_pd_sq:
	free_dma_memory(dev, graid_mem_kaddrs->pd_sqs, graid_mem_paddrs->pd_sqs,
			PDSQBytes, PDQ_COUNT);
out_free_ipc:
	kfree((void*)graid_mem_kaddrs->ipc);
out_free:
	if (graid_bar1_kaddrs) {
		vfree(graid_bar1_kaddrs);
		graid_bar1_kaddrs = NULL;
	}
	if (graid_bar1_paddrs) {
		vfree(graid_bar1_paddrs);
		graid_bar1_paddrs = NULL;
	}
	if (graid_mem_paddrs) {
		vfree(graid_mem_paddrs);
		graid_mem_paddrs = NULL;
	}
	if (graid_mem_kaddrs) {
		vfree(graid_mem_kaddrs);
		graid_mem_paddrs = NULL;
	}
	return -ENOMEM;
}

static void graid_free_page_resources(void)
{
	int i, pg;
	if (!res)
		return;

	for (i = 0; i < res->dma_pool.numa_nr; ++i) {
		int node = res->dma_pool.numa_nodes[i];
		if (res->page_pools[node].ptrs)
			vfree(res->page_pools[node].ptrs);
		if (res->dma_pool.kern_page_base[node]) {
			vm_unmap_ram(res->dma_pool.kern_page_base[node],
				     res->dma_pool.kern_alloc_pgnr);
			for (pg = 0; pg < res->dma_pool.kern_alloc_pgnr; ++pg)
				__free_pages((res->kern_page_lists[node])[pg], 0);
			vfree(res->kern_page_lists[node]);
		}
		if (res->dma_pool.gpu_page_base[node]) {
			vm_unmap_ram(res->dma_pool.gpu_page_base[node],
				     res->dma_pool.gpu_alloc_pgnr);
			for (pg = 0; pg < res->dma_pool.gpu_alloc_pgnr; ++pg)
				__free_pages((res->gpu_page_lists[node])[pg], 0);
			vfree(res->gpu_page_lists[node]);
		}
	}
	vfree(res);
	res = NULL;
}

static void* graid_alloc_dma_pages_node(int node, u64 page_nr, struct page ***page_list)
{
	void *ret;
	int cur_pgnr;

	u64 page_list_size = sizeof(struct page*) * page_nr;
	*page_list = vzalloc_node(page_list_size, node);
	if (!*page_list)
		return NULL;

	for (cur_pgnr = 0; cur_pgnr < page_nr; ++cur_pgnr) {
		(*page_list)[cur_pgnr] = alloc_pages_node(node, GFP_KERNEL, 0);
		if (unlikely(!(*page_list)[cur_pgnr]))
			goto free_out;
	}

#ifdef USE_OLD_VM_MAP_RAM
	ret = vm_map_ram(*page_list, page_nr, node, PAGE_KERNEL);
#else
	ret = vm_map_ram(*page_list, page_nr, node);
#endif
	if (!ret)
		goto free_out;

	return ret;

free_out:
	while (--cur_pgnr >= 0)
		__free_pages((*page_list)[cur_pgnr], 0);
	vfree(*page_list);
	*page_list = NULL;
	return NULL;
}

static int graid_setup_dma_pool_node(int node, u32 alloc_pgnr, u32 pgnr,
				     struct page ***page_list,
				     PageData_t **page_base, uint64_t **phy_tbl)
{
	int i;
	*page_base = (PageData_t*)graid_alloc_dma_pages_node(node, alloc_pgnr, page_list);
	if (!*page_base || (((uint64_t)*page_base) & 0xFFF))
		return -ENOMEM;
	*phy_tbl = (uint64_t*)((*page_base) + pgnr);
	for (i = 0; i < pgnr; ++i) {
		(*phy_tbl)[i] = page_to_pfn((*page_list)[i]) << PAGE_SHIFT;
	}
	return 0;
}

static int graid_alloc_page_resources(void)
{
	int i, rc, node;
	u32 kern_table_pgnr, gpu_table_pgnr;

	res = vzalloc(sizeof(struct PageResources));
	if (!res)
		return -ENOMEM;

	for_each_node_with_cpus(node) {
		if (node >= GRAID_MAX_NUMA)
			goto out_free;
		res->dma_pool.numa_nodes[res->dma_pool.numa_nr++] = node;
	}

	res->dma_pool.kbuf_ppn = GRAID_KBUF_PAGENR / res->dma_pool.numa_nr;
	res->dma_pool.pdprp_ppn = DIV_ROUND_UP(GRAID_PDPRP_PAGENR, res->dma_pool.numa_nr);
	res->dma_pool.pdprp_ppn = roundup(res->dma_pool.pdprp_ppn, GRAID_PDPRP_PAGENR_PER_CMD);
	res->dma_pool.parity_ppn = DIV_ROUND_UP(GRAID_PARITY_PAGENR, res->dma_pool.numa_nr);
	res->dma_pool.parity_ppn = roundup(res->dma_pool.parity_ppn, GRAID_PARITY_PAGENR_PER_CMD);
	res->dma_pool.rebuild_ppn = DIV_ROUND_UP(GRAID_REBUILD_PAGENR, res->dma_pool.numa_nr);
	res->dma_pool.rebuild_ppn = roundup(res->dma_pool.rebuild_ppn, GRAID_REBUILD_PAGENR_PER_CMD);

	res->dma_pool.kern_pgnr = res->dma_pool.kbuf_ppn;
	res->dma_pool.gpu_pgnr  = res->dma_pool.pdprp_ppn +
				  res->dma_pool.parity_ppn +
				  res->dma_pool.rebuild_ppn;
	kern_table_pgnr = DIV_ROUND_UP(res->dma_pool.kern_pgnr, PAGE_SIZE / sizeof(uint64_t));
	gpu_table_pgnr  = DIV_ROUND_UP(res->dma_pool.gpu_pgnr,  PAGE_SIZE / sizeof(uint64_t));
	res->dma_pool.kern_alloc_pgnr = res->dma_pool.kern_pgnr + kern_table_pgnr;
	res->dma_pool.gpu_alloc_pgnr  = res->dma_pool.gpu_pgnr  + gpu_table_pgnr;

	for_each_node_with_cpus(node) {
		rc = graid_setup_dma_pool_node(node, res->dma_pool.kern_alloc_pgnr,
					       res->dma_pool.kern_pgnr,
					       &res->kern_page_lists[node],
					       &res->dma_pool.kern_page_base[node],
					       &res->dma_pool.kern_paddr_tbls[node]);
		if (rc)
			goto out_free;
		rc = graid_setup_dma_pool_node(node, res->dma_pool.gpu_alloc_pgnr,
					       res->dma_pool.gpu_pgnr,
					       &res->gpu_page_lists[node],
					       &res->dma_pool.gpu_page_base[node],
					       &res->dma_pool.gpu_paddr_tbls[node]);
		if (rc)
			goto out_free;
	}

	res->page_pool_size = res->dma_pool.kbuf_ppn << 1;
	while (res->page_pool_size & (res->page_pool_size - 1))
		res->page_pool_size &= res->page_pool_size - 1;
	if (res->page_pool_size <= res->dma_pool.kbuf_ppn)
		res->page_pool_size <<= 1;
	res->page_pool_mask = res->page_pool_size - 1;
	for_each_node_with_cpus(node) {
		res->page_pools[node].ptrs = vmalloc_node(sizeof(struct ppptr) * res->page_pool_size, node);
		if (!res->page_pools[node].ptrs)
			goto out_free;

		for (i = 0; i < res->page_pool_size; ++i)
			atomic64_set(&res->page_pools[node].ptrs[i].kptr,
				     (i < res->dma_pool.kbuf_ppn) ? (uint64_t)(res->dma_pool.kern_page_base[node] + i) : 0);

		atomic_set(&res->page_pools[node].head, 0);
		atomic_set(&res->page_pools[node].tail, res->dma_pool.kbuf_ppn);
		atomic_set(&res->page_pools[node].free, res->dma_pool.kbuf_ppn);
	}

	res->gpu_page_ppn = 0;
	res->gpu_page_pool_size = 0;
	res->gpu_page_pool_mask = 0;

	return 0;

out_free:
	graid_free_page_resources();
	return -ENOMEM;
}

int init_dma_pools(struct device *dev)
{
	int rc;

	rc = graid_alloc_page_resources();
	if (rc)
		return -ENOMEM;

	rc = graid_alloc_queue_buffers(dev);
	if (rc)
		graid_free_page_resources();

	return rc;
}

void graid_destroy_gpu_page_pool(void)
{
	int n, node;
	int numa_nr = res->dma_pool.numa_nr;

	if (!graid_bar1_kaddrs->gpu_pages[0])
		return;

	for (n = 0; n < numa_nr; ++n) {
		node = res->dma_pool.numa_nodes[n];
		if (res->gpu_page_pools[node].ptrs) {
			if (atomic_read(&res->gpu_page_pools[node].free) != res->gpu_page_ppn) {
				pr_err("%s: GPU Page pool still in use on node: %d free=%u, ppn=%u\n",
				       __func__, node, atomic_read(&res->gpu_page_pools[node].free),
				       res->gpu_page_ppn);
				return;
			}
			vfree(res->gpu_page_pools[node].ptrs);
			res->gpu_page_pools[node].ptrs = NULL;
			atomic_set(&res->gpu_page_pools[node].head, 0);
			atomic_set(&res->gpu_page_pools[node].tail, 0);
			atomic_set(&res->gpu_page_pools[node].free, 0);
		}
	}
	res->gpu_page_ppn = 0;
	res->gpu_page_pool_size = 0;
	res->gpu_page_pool_mask = 0;
}

void graid_get_gpu_page_addr(int n, int i, atomic64_t *kaddr, atomic64_t *paddr)
{
	int idx = (n * res->gpu_page_ppn) + i;
	int gpg_idx = idx >> (GPU_PAGE_SHIFT - PAGE_SHIFT);
	idx &= (1u << (GPU_PAGE_SHIFT - PAGE_SHIFT)) - 1;
	atomic64_set(kaddr, graid_bar1_kaddrs->gpu_pages[gpg_idx] + (idx << PAGE_SHIFT));
	atomic64_set(paddr, graid_bar1_paddrs->gpu_pages[gpg_idx] + (idx << PAGE_SHIFT));
}

int graid_create_gpu_page_pool(void)
{
	int n, node, i;
	int numa_nr = res->dma_pool.numa_nr;

	if (!graid_bar1_kaddrs->gpu_pages[0])
		return 0;

	res->gpu_page_ppn = (GPG_COUNT << (GPU_PAGE_SHIFT - PAGE_SHIFT)) / res->dma_pool.numa_nr;
	res->gpu_page_pool_size = res->gpu_page_ppn << 1;
	while (res->gpu_page_pool_size & (res->gpu_page_pool_size - 1))
		res->gpu_page_pool_size &= res->gpu_page_pool_size - 1;
	if (res->gpu_page_pool_size <= res->gpu_page_ppn)
		res->gpu_page_pool_size <<= 1;
	res->gpu_page_pool_mask = res->gpu_page_pool_size - 1;

	for (n = 0; n < numa_nr; ++n) {
		node = res->dma_pool.numa_nodes[n];
		res->gpu_page_pools[node].ptrs = vmalloc_node(sizeof(struct ppptr) * res->gpu_page_pool_size, node);
		if (!res->gpu_page_pools[node].ptrs)
			goto out_free;

		for (i = 0; i < res->gpu_page_pool_size; ++i) {
			if (i < res->gpu_page_ppn) {
				graid_get_gpu_page_addr(n, i,
							&res->gpu_page_pools[node].ptrs[i].kptr,
							&res->gpu_page_pools[node].ptrs[i].pptr);
			} else {
				atomic64_set(&res->gpu_page_pools[node].ptrs[i].kptr, 0);
				atomic64_set(&res->gpu_page_pools[node].ptrs[i].pptr, 0);
			}
		}

		atomic_set(&res->gpu_page_pools[node].head, 0);
		atomic_set(&res->gpu_page_pools[node].tail, res->gpu_page_ppn);
		atomic_set(&res->gpu_page_pools[node].free, res->gpu_page_ppn);
	}
	return 0;

out_free:
	graid_destroy_gpu_page_pool();
	return -ENOMEM;
}

void graid_unmap_bar1(void)
{
	int i;
	for (i = 0; i < PDQ_COUNT; ++i) {
		graid_bar1_kaddrs->pd_sqs[i] = 0;
		graid_bar1_paddrs->pd_sqs[i] = 0;
		graid_bar1_kaddrs->pd_cqs[i] = 0;
		graid_bar1_paddrs->pd_cqs[i] = 0;
	}
	for (i = 0; i < DGQ_COUNT; ++i) {
		graid_bar1_kaddrs->dg_sqs[i] = 0;
		graid_bar1_paddrs->dg_sqs[i] = 0;
		graid_bar1_kaddrs->dg_cqs[i] = 0;
		graid_bar1_paddrs->dg_cqs[i] = 0;
	}
	for (i = 0; i < DB_COUNT; ++i) {
		graid_bar1_kaddrs->dbs[i] = 0;
		graid_bar1_paddrs->dbs[i] = 0;
	}
	graid_destroy_gpu_page_pool();
	for (i = 0; i < GPG_COUNT; ++i) {
		graid_bar1_kaddrs->gpu_pages[i] = 0;
		graid_bar1_paddrs->gpu_pages[i] = 0;
	}
	bar1_mapped = 0;
}

void cleanup_dma_pools(struct device *dev)
{
	graid_free_page_resources();

	if (graid_mem_kaddrs) {
		kfree((void*)graid_mem_kaddrs->ipc);
		free_dma_memory(dev, graid_mem_kaddrs->pd_sqs,
				graid_mem_paddrs->pd_sqs,
				PDSQBytes, PDQ_COUNT);
		free_dma_memory(dev, graid_mem_kaddrs->pd_cqs,
				graid_mem_paddrs->pd_cqs,
				PDCQBytes, PDQ_COUNT);
		free_dma_memory(dev, graid_mem_kaddrs->dg_sqs,
				graid_mem_paddrs->dg_sqs,
				DGSQBytes, DGQ_COUNT);
		free_dma_memory(dev, graid_mem_kaddrs->dg_cqs,
				graid_mem_paddrs->dg_cqs,
				DGCQBytes, DGQ_COUNT);
		free_dma_memory(dev, graid_mem_kaddrs->dbs,
				graid_mem_paddrs->dbs,
				PAGE_SIZE, DB_COUNT);
		vfree(graid_mem_kaddrs);
		graid_mem_kaddrs = NULL;
	}

	if (graid_mem_paddrs) {
		vfree(graid_mem_paddrs);
		graid_mem_paddrs = NULL;
	}

	if (graid_bar1_kaddrs) {
		vfree(graid_bar1_kaddrs);
		graid_bar1_kaddrs = NULL;
	}

	if (graid_bar1_paddrs) {
		vfree(graid_bar1_paddrs);
		graid_bar1_paddrs = NULL;
	}
}

void _graid_alloc_page(struct GRAID_Page_Pool *pool, uint32_t pool_mask, void **ret_kptr, uint64_t *ret_pptr)
{
	unsigned long flags;
	u32 pos, cnt;
	u64 kptr, pptr;
	atomic64_t *p_kptr, *p_pptr;

	local_irq_save(flags);
	preempt_disable();

	pos = (atomic_inc_return(&pool->head) - 1) & pool_mask;
	p_kptr = &pool->ptrs[pos].kptr;
	kptr = atomic64_add_return(0, p_kptr);
	cnt = 0;
	while (kptr == 0) {
		kptr = atomic64_add_return(0, p_kptr);
		if (++cnt == 1000)
			pr_info("%s: Waited 1000 cycles for kptr, pptr=%llx\n", __func__, (u64)ret_pptr);
	}
	atomic64_set(p_kptr, 0);
	*ret_kptr = (void*)kptr;

	if (ret_pptr) {
		p_pptr = &pool->ptrs[pos].pptr;
		pptr = atomic64_add_return(0, p_pptr);
		cnt = 0;
		while (pptr == 0) {
			pptr = atomic64_add_return(0, p_pptr);
			if (++cnt == 1000)
				pr_info("%s: Waited 1000 cycles for pptr, pptr=%llx\n", __func__, (u64)ret_pptr);
		}
		atomic64_set(p_pptr, 0);
		*ret_pptr = pptr;
	}

	local_irq_restore(flags);
	preempt_enable();
}

int graid_alloc_pages(int node, u32 page_nr, void **page_kptr_list)
{
	u32 i;

	if (unlikely(!res || !res->page_pools[node].ptrs)) {
		pr_err("%s: Tried to allocate without pool node=%d\n", __func__, node);
		return -ENOMEM;
	}
	if (unlikely(((int)page_nr) > atomic_read(&res->page_pools[node].free))) {
		pr_debug("Not enough pages want=%u free=%d\n",
			 page_nr, atomic_read(&res->page_pools[node].free));
		return -ENOMEM;
	}
	if (unlikely(atomic_sub_return(page_nr, &res->page_pools[node].free) < 0)) {
		atomic_add(page_nr, &res->page_pools[node].free);
		pr_debug("Out of pages want=%u\n", page_nr);
		return -ENOMEM;
	}

	mb();
	for (i = 0; i < page_nr; ++i) {
		_graid_alloc_page(&res->page_pools[node],
				  res->page_pool_mask,
				  page_kptr_list + i, NULL);
		pr_debug("Allocated from Page pool kptr=%llx\n", (u64)page_kptr_list[i]);
	}

	pr_debug("Allocated from Page pool pgnr=%u\n", page_nr);
	return 0;
}

int graid_alloc_gpu_page(int node, void **page_kptr, uint64_t *page_pptr)
{
	if (!res || res->gpu_page_pool_size == 0) {
		return -ENOMEM;
	}
	if (unlikely(!res->gpu_page_pools[node].ptrs)) {
		pr_err("%s: Tried to allocate without GPU Page pool node=%d\n", __func__, node);
		return -ENOMEM;
	}
	if (unlikely(atomic_read(&res->gpu_page_pools[node].free) < 1)) {
		return -ENOMEM;
	}
	if (unlikely(atomic_sub_return(1, &res->gpu_page_pools[node].free) < 0)) {
		atomic_add(1, &res->gpu_page_pools[node].free);
		return -ENOMEM;
	}
	mb();

	_graid_alloc_page(&res->gpu_page_pools[node],
			  res->gpu_page_pool_mask,
			  page_kptr, page_pptr);

	pr_debug("Allocated from GPU Page pool\n");
	return 0;
}

static inline
int graid_kptr_in_pool(int node, u64 kptr)
{
	return kptr >= (u64)GRAID_KBUF_PAGE_START(&res->dma_pool, node) &&
	       kptr < (u64)GRAID_KBUF_PAGE_END(&res->dma_pool, node);
}

u32 _graid_free_page(u64 kptr, u64 pptr, struct GRAID_Page_Pool *page_pool, u32 pool_mask)
{
	u32 pos, cnt;
	unsigned long flags;

	local_irq_save(flags);
	preempt_disable();

	pos = (atomic_inc_return(&page_pool->tail) - 1) & pool_mask;
	cnt = 0;
	while (atomic64_cmpxchg_relaxed(&page_pool->ptrs[pos].kptr, 0, kptr) != 0) {
		if (++cnt == 1000)
			pr_info("%s: Waited 1000 cycles\n", __func__);
	}

	if (pptr) {
		cnt = 0;
		while (atomic64_cmpxchg_relaxed(&page_pool->ptrs[pos].pptr, 0, pptr) != 0) {
			if (++cnt == 1000)
				pr_info("%s: Waited 1000 cycles\n", __func__);
		}
	}

	local_irq_restore(flags);
	preempt_enable();

	wmb();
	atomic_inc(&page_pool->free);

	return pos;
}

void graid_free_page(int node, void *page_kptr)
{
	u32 pos;
	u64 kptr = (u64)page_kptr;

	if (unlikely(!res || !res->page_pools[node].ptrs)) {
		pr_err("%s: Tried to free address without pool: %016llx, node=%u\n", __func__, kptr, node);
		return;
	}
	if (unlikely(!graid_kptr_in_pool(node, kptr))) {
		pr_err("%s: Tried to free invalid address: %016llx\n", __func__, kptr);
		return;
	}

	pos = _graid_free_page(kptr, 0, &res->page_pools[node], res->page_pool_mask);

	pr_debug("Freed page with pos=%u, kptr=%016llx, free=%u\n", pos, kptr,
		 atomic_read(&res->page_pools[node].free));
}

void graid_free_pages(int node, u32 page_nr, void **page_kptr_list)
{
	u32 i, pos;

	if (unlikely(!res || !res->page_pools[node].ptrs)) {
		pr_err("Tried to free pages without pool, node=%u\n", node);
		return;
	}

	for (i = 0; i < page_nr; ++i) {
		u64 kptr = (u64)page_kptr_list[i];
		if (unlikely(!graid_kptr_in_pool(node, kptr))) {
			pr_err("%s: Tried to free invalid address: %016llx\n", __func__, kptr);
			continue;
		}
		pos = _graid_free_page(kptr, 0, &res->page_pools[node], res->page_pool_mask);
		pr_debug("Freed page with pos=%u, kptr=%016llx, free=%u, head=%u, tail=%u\n", pos, kptr,
			 atomic_read(&res->page_pools[node].free),
			 atomic_read(&res->page_pools[node].head),
			 atomic_read(&res->page_pools[node].tail));
	}
}

void graid_free_gpu_page(int node, void *page_kptr, u64 page_pptr)
{
	u32 pos;
	u64 kptr = (u64)page_kptr;

	if (unlikely(!res || !res->gpu_page_pools[node].ptrs)) {
		pr_err("%s: Tried to free address without GPU page pool: %016llx, node=%u\n", __func__, kptr, node);
		return;
	}
	if (unlikely(!graid_bar1_p2k(page_pptr))) {
		pr_err("%s: Tried to free invalid address: %016llx\n", __func__, kptr);
		return;
	}

	pos = _graid_free_page(kptr, page_pptr, &res->gpu_page_pools[node], res->gpu_page_pool_mask);

	pr_debug("Freed GPU page with pos=%u, kptr=%016llx, free=%u, head=%u, tail=%u\n", pos, kptr,
		 atomic_read(&res->gpu_page_pools[node].free),
		 atomic_read(&res->gpu_page_pools[node].head),
		 atomic_read(&res->gpu_page_pools[node].tail));
}

dma_addr_t graid_map_page(struct page *page)
{
	return page_to_phys(page);
}

long graid_ioctl_get_dma_pool(void __user* pool)
{
	if (!res)
		return -ENOMEM;
	if (copy_to_user(pool, &res->dma_pool, sizeof(res->dma_pool)))
		return -EFAULT;
	return 0;
}

int graid_ioctl_get_mem_addrs(void __user *arg)
{
	BUG_ON(!graid_mem_paddrs);
	if (copy_to_user(arg, graid_mem_paddrs, sizeof(struct graid_mem_addrs)))
		return -EFAULT;
	return 0;
}

static int _graid_get_bar1_kaddr(uint64_t paddr, uint64_t *kaddr, const char *type, int idx)
{
	void *ka;
	if (paddr == 0)
		return 0;

	ka = graid_bar1_p2k(paddr);
	if (ka == NULL) {
		pr_info("%s: Failed to get kernel address of %s[%d]:%016llx\n",
			__func__, type, idx, paddr);
		return -EFAULT;
	}
	*kaddr = (uint64_t)ka;
	pr_debug("%s: %s[%d]:%016llx@%016llx\n",
		 __func__, type, idx, paddr, *kaddr);
	return 0;
}

int graid_ioctl_dma_set_bar1(void __user *arg)
{
	int i;

	if (bar1_mapped)
		return -EBUSY;

	if (copy_from_user(graid_bar1_paddrs, arg, sizeof(struct graid_mem_addrs))) {
		pr_err("%s: Failed to copy request from user\n", __func__);
		return -EFAULT;
	}

	bar1_mapped = 1;
	for (i = 0; i < PDQ_COUNT; ++i) {
		if (_graid_get_bar1_kaddr(graid_bar1_paddrs->pd_sqs[i],
					  graid_bar1_kaddrs->pd_sqs + i,
					  "PD_SQ", i))
			goto err_out;
		if (_graid_get_bar1_kaddr(graid_bar1_paddrs->pd_cqs[i],
					  graid_bar1_kaddrs->pd_cqs + i,
					  "PD_CQ", i))
			goto err_out;
	}
	for (i = 0; i < DGQ_COUNT; ++i) {
		if (_graid_get_bar1_kaddr(graid_bar1_paddrs->dg_sqs[i],
					  graid_bar1_kaddrs->dg_sqs + i,
					  "DG_SQ", i))
			goto err_out;
		if (_graid_get_bar1_kaddr(graid_bar1_paddrs->dg_cqs[i],
					  graid_bar1_kaddrs->dg_cqs + i,
					  "DG_CQ", i))
			goto err_out;
	}

	for (i = 0; i < DB_COUNT; ++i) {
		if (_graid_get_bar1_kaddr(graid_bar1_paddrs->dbs[i],
					  graid_bar1_kaddrs->dbs + i,
					  "DBS", i))
			goto err_out;
	}

	for (i = 0; i < GPG_COUNT; ++i) {
		if (_graid_get_bar1_kaddr(graid_bar1_paddrs->gpu_pages[i],
					  graid_bar1_kaddrs->gpu_pages + i,
					  "GPG", i))
			goto err_out;
	}

	if (graid_create_gpu_page_pool())
		goto err_out;

	return 0;

err_out:
	graid_unmap_bar1();
	return -EFAULT;
}

int graid_ioctl_dma_reset_bar1(void __user *arg)
{
	graid_unmap_bar1();
	return 0;
}

int graid_map_dmapool_pages(struct vm_area_struct *vma, int node, size_t start)
{
	if (!res || !res->kern_page_lists[node] || !res->gpu_page_lists[node])
		return -ENOMEM;

	if ((size_t)res->dma_pool.kern_page_base[node] == start) {
		return vm_map_pages_zero(vma, res->kern_page_lists[node], res->dma_pool.kern_alloc_pgnr);
	} else if ((size_t)res->dma_pool.gpu_page_base[node] == start) {
		return vm_map_pages_zero(vma, res->gpu_page_lists[node], res->dma_pool.gpu_alloc_pgnr);
	} else {
		return -EINVAL;
	}
}
