#ifndef __DMAPOOL_H_INCLUDED__
#define __DMAPOOL_H_INCLUDED__

#ifndef __KERNEL__
#include <stdint.h>
#else
#include <linux/types.h>
#endif

#define GRAID_MAX_IO_PAGES (PAGE_SIZE / sizeof(uint64_t))

#define NVMeMaxPDNR 32u
#define NVMeMaxPDQNR 64u
#define NVMeMinPDQDepthShift 8u
#define NVMeMaxPDQDepthShift 9u
#define NVMeMaxPDQDepth (1 << NVMeMaxPDQDepthShift)
#define NVMeMaxPDQDepthMask (NVMeMaxPDQDepth - 1)

#define NVMeMaxDGNR 4u
#define NVMeMaxDGQNR 128u
#define NVMeMaxDGQDepthShift 10u
#define NVMeMaxDGQDepth (1 << NVMeMaxDGQDepthShift)
#define NVMeMaxDGQDepthMask (NVMeMaxDGQDepth - 1)
#define GraidMaxDGInflightCMD (NVMeMaxDGQDepth - 8)

#define NVMeMaxVDNR 8u
#define NVMeMaxVDQDepth 250u

#define NVMeSQEShift 6u
#define NVMeSQESize 64u
#if (1u << NVMeSQEShift) != NVMeSQESize
#error "Incorrect NVMeSQEShift"
#endif
#define NVMeCQEShift 4u
#define NVMeCQESize 16u
#if (1u << NVMeCQEShift) != NVMeCQESize
#error "Incorrect NVMeCQEShift"
#endif

#define DGSQBytes (NVMeSQESize * NVMeMaxDGQDepth)
#define DGCQBytes (NVMeCQESize * NVMeMaxDGQDepth)
#define PDSQBytes (NVMeSQESize * NVMeMaxPDQDepth)
#define PDCQBytes (NVMeCQESize * NVMeMaxPDQDepth)

#define GraidTotalPDQNR (NVMeMaxPDNR * NVMeMaxPDQNR)
#define GraidTotalDGQNR (NVMeMaxDGNR * NVMeMaxDGQNR)
#define GraidTotalQPNR (GraidTotalPDQNR + GraidTotalDGQNR)

#define RCMD_ENT_NR_SHIFT 12
#define RCMD_ENT_NR (1u << RCMD_ENT_NR_SHIFT)
#define RCMD_ENT_MASK (RCMD_ENT_NR - 1u)
#define P2P_PGNR_PER_CMD 3

#define GPU_PAGE_SHIFT 16
#define GPU_PAGE_SIZE  (1llu << GPU_PAGE_SHIFT)
#define GPU_PAGE_MASK  (GPU_PAGE_SIZE - 1)
#define MaxBAR1PGNR (256 << (20 - GPU_PAGE_SHIFT))

#define MAX_REBUILD_DEPTH 16

typedef struct {
	uint8_t raw[4096];
} PageData_t;

#define GRAID_MAX_NUMA 128
#define GRAID_GPU_FAKE_NUMA 192

/* Total 2 GiB kernel page buffer */
#define GRAID_KBUF_PAGENR ((2llu << 30) >> 12)

/* Fixed PRP spaces for each RAID command on each PD */
#define GRAID_PDPRP_PAGENR_PER_CMD ((sizeof(uint64_t) * 32 * NVMeMaxPDNR) / 4096llu)
#define GRAID_PDPRP_PAGENR ((RCMD_ENT_NR + MAX_REBUILD_DEPTH) * GRAID_PDPRP_PAGENR_PER_CMD)

/* Fixed Old-Data/Parity buffer space for each RAID command */
#define GRAID_PARITY_PAGENR_PER_CMD 128llu
#define GRAID_PARITY_PAGENR (RCMD_ENT_NR * GRAID_PARITY_PAGENR_PER_CMD)

/* Fixed buffer space for rebuild/init, reserve 128K(32 pages) for each rebuild command on each PD */
#define GRAID_REBUILD_PAGENR_PER_CMD (NVMeMaxPDNR * 32llu)
#define GRAID_REBUILD_PAGENR (MAX_REBUILD_DEPTH * GRAID_REBUILD_PAGENR_PER_CMD)

struct GRAID_DMA_Pool {
	uint32_t numa_nr;
	uint32_t kern_pgnr;
	uint32_t gpu_pgnr;
	uint32_t kern_alloc_pgnr;
	uint32_t gpu_alloc_pgnr;
	uint32_t kbuf_ppn;
	uint32_t pdprp_ppn;
	uint32_t parity_ppn;
	uint32_t rebuild_ppn;
	uint8_t  numa_nodes[GRAID_MAX_NUMA];
	PageData_t *kern_page_base[GRAID_MAX_NUMA];
	PageData_t *gpu_page_base[GRAID_MAX_NUMA];
	uint64_t *kern_paddr_tbls[GRAID_MAX_NUMA];
	uint64_t *gpu_paddr_tbls[GRAID_MAX_NUMA];
};

struct GRAID_Addr_Offsets {
	int64_t kern_kptr_to_gptr[GRAID_MAX_NUMA];
	int64_t kern_kptr_to_uptr[GRAID_MAX_NUMA];
	int64_t gpu_kptr_to_gptr[GRAID_MAX_NUMA];
	int64_t gpu_kptr_to_uptr[GRAID_MAX_NUMA];
};

#define GRAID_DMAPOOL_DPTR(node, kptr) (((uint64_t)(kptr)) | (node + 1))
#define GRAID_DMAPOOL_KPTR(dptr) (void *)(((uint64_t)(dptr)) & (~((1llu << 12) - 1)))

#define GRAID_KBUF_PAGE_START(dma_pool, node) ((dma_pool)->kern_page_base[node] + 0)
#define GRAID_KBUF_PAGE_END(dma_pool, node) ((dma_pool)->kern_page_base[node] + (dma_pool)->kbuf_ppn)

#endif
